ORDER = 9


def renew_grid():
    global rows
    global cols
    rows = [chr(i+97) if i < 26 else chr(i//26 + 96) + chr((i % 26) + 97) for i in range(ORDER)]
    cols = [str(i+1) for i in range(ORDER)]


renew_grid()



import os
import re
import subprocess
from src import settings


def solve(path, tex):
    limboole = os.environ["limboole"]
    satisfiable = True
    i = 0
    while satisfiable:
        p = subprocess.Popen(f"{limboole} -s {path}", stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, shell=True)
        (output, error) = p.communicate()
        p.wait()

        if "UNSATISFIABLE formula" in str(output):
            satisfiable = False
            continue

        solution, prefills = extract_solution(str(output))
        i += 1
        print(f"Solution #{i}:")
        if tex:
            texify(solution)
        else:
            prettify(solution)

        prefills = f" & !({' & '.join(prefills)})"
        with open(path, "a") as file:
            file.write(prefills)

    if i == 0:
        print("Sudoku is unsatisfiable.")
    elif i == 1:
        print("Sudoku is uniquely solvable. No further solutions exist.")
    else:
        print(f"No further solutions exist. Total number of solutions: {i}")


def solve_test():
    limboole = os.environ["limboole"]
    path = os.environ["outpath"]
    p = subprocess.Popen(f"{limboole} -s {path}", stdout=subprocess.PIPE, shell=True)
    (output, error) = p.communicate()
    p.wait()

    solution, prefills = extract_solution(str(output))


def extract_solution(output):
    output = output.replace(r"\n", "\n")
    solution = ["."] * settings.ORDER**2
    prefills = []

    matches = re.finditer(r"(?P<i>[a-z]+)(?P<j>\d+)_(?P<k>\d+) = 1", output, re.MULTILINE)
    for match in matches:
        i = match.group("i")
        j = int(match.group("j"))
        k = match.group("k")
        solution[(j - 1) + (settings.rows.index(i)) * settings.ORDER] = k
        prefills.append(f"{i}{j}_{k}")

    return solution, prefills


def texify(solution):
    solution = [solution[i::9] for i in range(9)]
    for s in solution:
        print("|" + "|".join(s) + "|")
    print()


def prettify(solution):
    c = 1 if settings.ORDER < 10 else 2
    order = int(settings.ORDER**(.5))
    header = "     " + " | ".join(["-".join(["-" * c] * order)] * order)
    print()
    print(header)

    for i in range(settings.ORDER):
        line = solution[i*settings.ORDER:(i+1)*settings.ORDER]
        offset = 0
        line = [x if len(x) == c else f" {x}" for x in line]
        for j in range(1, settings.ORDER):
            if j % order == 0:
                line.insert(j+offset, " ")
                offset += 1
        line = " ".join(line)
        print(f"   | {line}")

        if i in [j for j in range(settings.ORDER - 1) if (j + 1) % order == 0]:
            print("   -")
    print()



import re
from itertools import product
from src import settings


def is_atomic(clause):
    if isinstance(clause, list):
        return False
    if "&" in clause or "|" in clause:
        return False
    return True


def make_atomic(clause):
    clause.replace("(", "")
    clause.replace(")", "")
    clause = re.sub(r"\s", "", clause)
    atoms = re.split(r"&|\|", clause)
    return atoms


def deconstruct(atom):
    row = re.sub(r"[^a-z]", "", atom)
    col, val = re.sub(r"[a-z]",  "", atom).split("_")
    return row, col, val


def is_allowed(left, right):
    atoms_l = make_atomic(left)
    atoms_r = make_atomic(right)

    for a_l in atoms_l:
        row, col, val = deconstruct(a_l)
        for a in atoms_r:
            r, c, v = deconstruct(a)
            if (row == r and val == v and col != c) or \
               (col == c and val == v and row != r) or \
               (row == r and col == c and val != v):
                return False
    return True


def and_clause(left, right):
    if not isinstance(left, list):
        left = [left]
    if not isinstance(right, list):
        right = [right]
    variants = list(product(range(len(left)), range(len(right))))
    finished = []
    for v in variants:
        variant = [left[v[0]], right[v[1]]]
        if "False" in variant:
            variant = ["False"]
        elif "True" in variant:
            variant.remove("True")
        elif not is_allowed(*variant):
            continue
        if len(variant) > 1:
            finished.append(" & ".join(variant))
        else:
            finished.append(variant[0])
    return list(set(finished))


def simple_and(ks):
    ks = set(ks)
    ks.discard("True")
    if not len(ks):
        return "True"
    ks = list(ks)
    if len(ks) > 1:
        return " & ".join(ks)
    return ks[0]


def or_clause(ks):
    if len(ks):
        ks = list(set(ks))
        if "False" in ks:
            ks.remove("False")
        return " | ".join(["(" + k + ")" if not is_atomic(k) and not k[0] == "(" else k for k in ks])
    return "ERR"


def grouped(clause):
    if is_atomic(clause):
        return clause
    return f"({clause})" if len(clause) else None


def new_buffer(size=None):
    if not size:
        size = settings.ORDER
    return [[] for _ in range(size)]


def equalize(left, right):
    while len(left) < len(right):
        left.append(["False"])
    while len(right) < len(left):
        right.append(["False"])
    return left, right


def reduce(propositions, mode="and"):
    output = []
    for truth in propositions:
        t = []
        for option in truth:
            option = [o for o in option if "False" not in o]
            if option:
                if len(option) > 1:
                    t.append(or_clause(option))
                else:
                    t.append(or_clause(option))
        output.append(grouped(or_clause(t)))
    if mode == "and":
        return simple_and(output)
    elif mode == "or":
        return or_clause(output)



from itertools import product
from src.helpers import and_clause, equalize


def operate(func):
    def inner(left, right):
        value_map, max_len = func(len(left), len(right))
        left, right = equalize(left, right)
        values = []
        for i in range(max_len):
            vals = []
            for pair in value_map[i]:
                vals.extend(and_clause(left[pair[0]], right[pair[1]]))
            vals[:] = [v for v in vals if v != "False"]
            vals = vals if vals else ["False"]
            values.append(vals)
        return values
    return inner


@operate
def ADD(left, right):
    options = list(product(range(left), range(right)))
    legal_pairs = []
    for i in range(left + right):
        vals = []
        for pair in options:
            value = pair[0] + pair[1]
            if value+1 == i:
                vals.append(pair)
        legal_pairs.append(vals)
    return legal_pairs, left+right


@operate
def SUB(left, right):
    options = list(product(range(left), range(left)))
    legal_pairs = []
    for i in range(left):
        vals = []
        for pair in options:
            value = pair[0] - pair[1]
            if value-1 == i:
                vals.append(pair)
        legal_pairs.append(vals)
    return legal_pairs, left


@operate
def MULT(left, right):
    options = list(product(range(left), range(right)))
    legal_pairs = []
    for i in range(left * right):
        vals = []
        for pair in options:
            value = (pair[0]+1) * (pair[1]+1)
            if value == i+1:
                # print(i-1, pair[0]+1, pair[1]+1)
                vals.append(pair)
        legal_pairs.append(vals)
    return legal_pairs, left*right


def OR(left, right):
    left, right = equalize(left, right)
    buffer = []
    for i in range(len(left)):
        combined = set([*left[i], *right[i]])
        if len(combined) > 1:
            combined.discard("False")
        buffer.append(list(combined))
    return buffer



import re
from src import settings


def modify(func):
    def inner(field):
        row = re.sub("[^a-z]", "", field)
        col = re.sub("[a-z]", "", field)
        return func(row, col)
    return inner


@modify
def NORTH(row, col):
    if row == "a":
        return ["ERR"]
    return [settings.rows[settings.rows.index(row) - 1] + col]


@modify
def SOUTH(row, col):
    if row == settings.rows[-1]:
        return ["ERR"]
    return [settings.rows[settings.rows.index(row) + 1] + col]


@modify
def EAST(row, col):
    if col == str(settings.ORDER):
        return ["ERR"]
    return [row + settings.cols[settings.cols.index(col) + 1]]


@modify
def WEST(row, col):
    if col == settings.rows[-1]:
        return ["ERR"]
    return [row + settings.cols[settings.cols.index(col) - 1]]


@modify
def NE(row, col):
    if row == "a" or col == str(settings.ORDER):
        return ["ERR"]
    return [settings.rows[settings.rows.index(row) - 1] + settings.cols[settings.cols.index(col) + 1]]


@modify
def SE(row, col):
    if row == settings.rows[-1] or col == str(settings.ORDER):
        return ["ERR"]
    return [settings.rows[settings.rows.index(row) + 1] + settings.cols[settings.cols.index(col) + 1]]


@modify
def NW(row, col):
    if row == "a" or col == "1":
        return ["ERR"]
    return [settings.rows[settings.rows.index(row) - 1] + settings.cols[settings.cols.index(col) - 1]]


@modify
def SW(row, col):
    if row == settings.rows[-1] or col == "1":
        return ["ERR"]
    return [settings.rows[settings.rows.index(row) + 1] + settings.cols[settings.cols.index(col) - 1]]


def HORIZONTAL(field):
    return [*WEST(field), *EAST(field)]


def VERTICAL(field):
    return [*NORTH(field), *SOUTH(field)]


def ORTHO(field):
    return [*HORIZONTAL(field), *VERTICAL(field)]


def ANY(field):
    return [*HORIZONTAL(field), *VERTICAL(field), *SE(field), *NE(field), *SW(field), *NW(field)]



from src import settings
from lark import Tree
from src.operators import ADD, SUB, MULT, OR
from src.comparators import EQ, NEQ, LT, LEQ, GT, GEQ, POR
from src.modifiers import NORTH, SOUTH, EAST, WEST, HORIZONTAL, VERTICAL, ORTHO, NE, SE, NW, SW, ANY
from src.prefixes import DISTINCT


class Compiler:
    def __init__(self, tree):
        self.tree = tree

    def get_propositions(self):
        # print(self.tree)
        return self.visit(self.tree)

    def visit(self, node):
        if isinstance(node, Tree):
            type, value, children = node.data.value, None, node.children
        else:
            type, value, children = node.type, node.value, []

        if type == "source":
            truths = []
            for child in children:
                truth = self.visit(child)
                truths.extend(truth)
            # print(truths)
            return truths

        elif type == "ORDER":
            settings.ORDER = int(value)**2
            settings.renew_grid()
            return [[["True"]]]

        elif type == "proposition":
            left, comparison, right = [self.visit(child) for child in children]
            results = []
            for i in left:
                for j in right:
                    results.append(comparison(i, j))
            return results

        elif type == "builtin":
            builtin, *args = [self.visit(child) for child in children]
            return builtin(args)

        elif type == "expression":
            if len(children) == 1:
                return self.visit(children[0])
            left, operator, right = [self.visit(child) for child in children]
            results = []
            for i in left:
                for j in right:
                    results.append(operator(i, j))
            return results

        elif type == "list":
            values = []
            for child in children:
                values.extend(self.visit(child))
            return values

        elif type == "CELL":
            if "." not in value:
                return [[[f"{value}_{i}"] for i in range(1, settings.ORDER+1)]]
            f, m = value.split(".")
            fields = self.modifier_map[m](f)
            return [[[f"{f}_{i}" if f != "ERR" else "False" for f in fields] for i in range(1, settings.ORDER+1)]]
        elif type == "NUMBER":
            num = [["False"] for i in range(1, int(value))]
            num.append(["True"])
            return [num]
        elif type == "OPERATOR":
            return self.operator_map[value]
        elif type == "COMPARATOR":
            return self.comparison_map[value]
        elif type == "PREFIX":
            return self.prefix_map[value]

    comparison_map = {
        "=": EQ,
        "<": LT,
        ">": GT,
        "<=": LEQ,
        ">=": GEQ,
        "!=": NEQ,
        "||": POR
    }

    operator_map = {
        "+": ADD,
        "-": SUB,
        "*": MULT,
        "|": OR
    }

    modifier_map = {
        "north": NORTH,
        "south": SOUTH,
        "east": EAST,
        "west": WEST,
        "horizontal": HORIZONTAL,
        "vertical": VERTICAL,
        "ortho": ORTHO,
        "ne": NE,
        "nw": NW,
        "se": SE,
        "sw": SW,
        "any": ANY
    }

    prefix_map = {
        "!!": DISTINCT
    }



import pathlib
import setuptools

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setuptools.setup(name='ssc-compiler',
                 version="1.0.0",

                 description='Describe generalized Sudokus through simple equations. Compile programs to SAT and solve them with limboole.',
                 long_description=README,
                 long_description_content_type="text/markdown",
                 keywords=['sudoku', 'sat', 'compiler', 'limboole'],
                 url='https://github.com/charludo/ssc',
                 author='Charlotte Hartmann Paludo',
                 author_email='charlotte.hartmann-paludo@stud.uni-due.de',
                 packages=setuptools.find_packages(),
                 include_package_data=True,
                 zip_safe=False,

                 entry_points={"console_scripts": ["ssc = src.ssc:run"]},
                 install_requires=[
                    "lark>=1.0.0",
                    "click>=8.0.3",
                    "humanize>=3.13.0"
                 ]
                 )



from itertools import product
from src.helpers import and_clause, new_buffer, equalize, reduce
from src import settings


def EQ(left, right):
    len_left = len(left)
    buffer = new_buffer(len_left)
    left, right = equalize(left, right)
    for i in range(len_left):
        options = and_clause(left[i], right[i])
        buffer[i] = options if len(options) else ["False"]
    return buffer


def NEQ(left, right):
    len_left = len(left)
    len_right = len(right)
    buffer = new_buffer(len_left)
    left, right = equalize(left, right)
    combinations = [(i, j) for i, j in product(range(len_left), range(len_right)) if i != j]
    for i, j in combinations:
        options = and_clause(left[i], right[j])
        buffer[i].extend(options if len(options) else ["False"])
    return buffer


def LT(left, right, offset=0):
    len_right = len(right)
    buffer = new_buffer(len_right)
    left, right = equalize(left, right)
    for i in range(1-offset, len_right):
        if right[i] != ["False"]:
            for j in range(0, i+offset):
                options = and_clause(left[j], right[i])
                buffer[j].extend(options if len(options) else ["False"])

    return [list(set(b)) for b in buffer]


def LEQ(left, right):
    return LT(left, right, offset=1)


def GT(left, right, offset=1):
    len_left = len(left)
    buffer = new_buffer(len_left)
    left, right = equalize(left, right)
    for i in range(0, len_left):
        if right[i] != ["False"]:
            for j in range(i+offset, settings.ORDER):
                options = and_clause(left[j], right[i])
                buffer[j].extend(options if len(options) else ["False"])

    return buffer


def GEQ(left, right):
    return GT(left, right, offset=0)


def POR(left, right):
    return [[reduce([left, right], mode="or")]]



import click
import timeit
import subprocess
from lark import Lark
from os import environ
from pathlib import Path
from datetime import timedelta
from humanize import naturalsize
from os.path import join, dirname, getsize
from src.compiler import Compiler
from src.interpreter import solve, solve_test
from src.base_rules import get_base_rules
from src.helpers import reduce


def get_parse_tree(code):
    file = open(join(dirname(__file__), "grammar.lark"), "r")
    parser = Lark(file, start="source")
    file.close()

    tree = parser.parse(code)
    return tree


@click.command()
@click.option("--base", "-b", help="specify file from which to read base rules")
@click.option("--interpret", "-i", help="apply limboole to compiled file and display result", is_flag=True)
@click.option("--limboole", "-l", help="specify and store absolute path to limboole executable", required=False, type=str)
@click.option("--minimal", "-m", help="do not append base rules", is_flag=True)
@click.option("--report", "-r", help="report on compilation and evaluation times. argument sets number of timing processes.", required=False, type=int)
@click.option("--tex", "-t", help="output solutions in LaTeX format", is_flag=True)
@click.option("--view", "-v", help="print compiled formula to console (excluding base rules)", is_flag=True)
@click.argument("filename")
def run(filename, view, tex, report, minimal, limboole, interpret, base):
    with open(filename, "r") as file:
        code = file.read()

    tree = get_parse_tree(code)
    compiler = Compiler(tree)

    propositions = compiler.get_propositions()
    # print(propositions)
    propositions = reduce(propositions)

    if view:
        print(propositions)

    if not minimal and not base:
        propositions += " & True & !ERR & " + get_base_rules()

    if base:
        with open(base, "r") as file:
            base_rules = file.read()
        propositions += " & True & !ERR & " + base_rules

    outpath = Path(filename).stem + ".sat"
    with open(outpath, "w") as file:
        file.write(propositions)

    print("done compiling.")
    print("compiled file size:      ", naturalsize(getsize(outpath)))

    if limboole:
        with open(".limboole", "w") as f:
            f.write(limboole)
        environ["limboole"] = limboole
    elif report or interpret:
        try:
            with open(".limboole", "r") as f:
                environ["limboole"] = f.read()
        except FileNotFoundError:
            p = subprocess.Popen("limboole -s ''", stdout=subprocess.PIPE, shell=True)
            (output, error) = p.communicate()
            p.wait()

            if "could no read" in str(output):
                environ["limboole"] = "limboole"
            else:
                print("limboole installation not found. Specify and store custom executable path with --limboole")
                return

    if report:
        environ["outpath"] = outpath
        # print("Performing timing measurements...")
        avg = timeit.timeit(compiler.get_propositions, number=report)/report
        print("Average compile time:    ", str(timedelta(seconds=avg)))
        avg = timeit.timeit(solve_test, number=report)/report
        print("Average evaluation time: ", str(timedelta(seconds=avg)))

    if interpret:
        solve(outpath, tex)


if __name__ == "__main__":
    run()



from src.helpers import new_buffer


def DISTINCT(*args):
    fields = [[v[0] for v in field] for arg in args[0] for field in arg]
    buffer = new_buffer()

    for i in range(len(fields)):
        left = fields[i]
        for j in range(i+1, len(fields)):
            right = fields[j]

            for v in range(len(left)):
                buffer[0].append(f"(!{left[v]} | !{right[v]})")

    buffer[0] = [" & ".join(buffer[0])]
    return [buffer]



from src import settings


def and_clause(ks):
    return " & ".join(ks)


def or_clause(ks):
    return " | ".join(ks)


def grouped(clause):
    return f"({clause})" if len(clause) else None


def get_base_rules():
    global ks
    global rows
    global cols
    global order
    ks = range(1, settings.ORDER+1)
    rows = settings.rows
    cols = settings.cols
    order = int(settings.ORDER**(.5))

    rs = and_clause([row(i) for i in rows])
    columns = and_clause([column(j) for j in cols])
    areas = and_clause([area(a) for a in ks])
    base_rules = and_clause([number_everywhere(), no_double_entries(), rs, columns, areas])
    return base_rules


def number_everywhere():
    return and_clause([grouped(or_clause([f"{i}{j}_{k}" for k in ks])) for i in rows for j in cols])


def no_double_entries():
    return and_clause([grouped(and_clause([f"(!{i}{j}_{k1} | !{i}{j}_{k2})" for k1 in ks for k2 in ks if k1 < k2])) for i in rows for j in cols])


def row(i):
    return and_clause([grouped(and_clause([f"(!{i}{j1}_{k} | !{i}{j2}_{k})" for k in ks])) for j1 in cols for j2 in cols if j1 < j2])


def column(j):
    return and_clause([grouped(and_clause([f"(!{i1}{j}_{k} | !{i2}{j}_{k})" for k in ks])) for i1 in rows for i2 in rows if i1 < i2])


def area(a):
    """
    area numbering scheme:
        1   2   3
        4   5   6
        7   8   9
    1 -> i=1, j=1
    2 -> i=1, j=4
    3 -> i=1, j=7

    4 -> i=4, j=1
    5 -> i=4, j=4
    6 -> i=4, j=7

    7 -> i=7, j=1
    8 -> i=7, j=4
    9 -> i=7, j=7
    """
    i_start = ((a - 1) // order) * order + 1
    j_start = ((a - 1) % order) * order + 1
    iv = range(i_start, i_start + order)
    jv = range(j_start, j_start + order)

    positions = [f"{rows[i-1]}{j}" for i in iv for j in jv]

    return and_clause([grouped(and_clause([f"(!{p1}_{k} | !{p2}_{k})" for k in ks])) for p1 in positions for p2 in positions if p1 != p2])
